#include <stdio.h>
#include "global.h"

/* memory allocation routines
 *
 * Adapted from Aztec user-supplied code by William C. Colley III; MUST
 * be used in place of original, *BUGGY* Aztec code!
 *
 * MSC Users note: can't use malloc since it does stack probes. Since Colley's
 *                 version does nice auditing, we'll grabcore at the beginning
 *		   and use Colley's code to do malloc program-wide (stack
 *		   probes in an interrupt context are evil since the SS may have
 *		   a different value...).  To remedy this, GLOBAL.H #defines
 *		   malloc, calloc, realloc, free to net_malloc, net_calloc,
 *		   net_realloc, and net_free.  grabcore #undef's malloc
 *		   for it's one massive malloc.  Don't add code after grabcore.
 *
 * Routines realloc() and calloc() were added; only minor format changes
 * made elsewhere.
 */

#ifndef	NULL
#define	NULL	(void *)0
#endif

union header {
	struct {
		union header *ptr;
		unsigned size;
	} s;
	long l;
};

typedef union header HEADER;
#define	NULLHDR	(HEADER *)NULL
static HEADER base, *allocp = NULLHDR;
static unsigned memfail;

/* Define ALLOCDEBUG for allocation recording. Note: works only with
 * small memory model
 */
#ifdef	ALLOCDEBUG
#define	NALLOC	150
/* Map of allocated memory chunks */
struct alloc {
	char *addr;
	unsigned size;
	long pc;
};
struct alloc alloc[NALLOC];
#endif

/* Allocate block of 'nb' bytes */
char *
malloc(nb)
unsigned nb;
{
	HEADER *morecore();
	register HEADER *p, *q;
	register unsigned nu;
	char i_state;

#ifdef	ALLOCDEBUG
	struct alloc *ap;
	unsigned short *ptr;
#endif

	i_state = disable();
	nu = ((nb + 3) >> 2) + 1;
	if ((q = allocp) == NULLHDR){
		base.s.ptr = allocp = q = &base;
		base.s.size = 1;
	}
	for (p = q->s.ptr; ; q = p, p = p->s.ptr){
		if (p->s.size >= nu){
			if (p->s.size == nu){
				q->s.ptr = p->s.ptr;
			} else {
				p->s.size -= nu;
				p += p->s.size;
				p->s.size = nu;
			}
			allocp = q;
			p->s.ptr = p;	/* for auditing */
#ifdef	ALLOCDEBUG
			for(ap = alloc;ap < &alloc[NALLOC];ap++){
				if(ap->addr == NULLCHAR){
					ap->addr = (char *)(p+1);
					ap->size = nb;
					ptr = &nb;
#ifdef	LARGECODE
					ap->pc = (long)ptr[-1] << 16 +
						(long)ptr[-2];
#else
					ap->pc = (long)ptr[-1];
#endif
					break;
				}
			}
#endif
			restore(i_state);
			return (char *)(p + 1);
		}
		if (p == allocp && (p = morecore(nu)) == NULLHDR){
			memfail++;
			restore(i_state);
			return NULLCHAR;
		}
	}
}

/* Get more memory from the system and put it on the heap */
HEADER *
morecore(nu)
unsigned nu;
{
	char *sbrk();
	register char *cp;
	register HEADER *up;

	if ((int)(cp = sbrk(nu << 2)) == -1)
		return NULLHDR;
	up = (HEADER *)cp;
	up->s.size = nu;
	up->s.ptr = up;	/* satisfy audit */
	free((char *)(up + 1));
	return allocp;
}
/* Put memory block back on heap */
free(blk)
char *blk;
{
	register HEADER *p, *q;
	unsigned short *ptr;
	char i_state;
#ifdef	ALLOCDEBUG
	struct alloc *ap;
#endif

	i_state = disable();
#ifdef	ALLOCDEBUG
	for(ap = alloc;ap < &alloc[NALLOC];ap++){
		if(ap->addr == blk){
			ap->addr = NULLCHAR;
			break;
		}
	}
#endif
	p = (HEADER *)blk - 1;
	/* Audit check */
	if(p->s.ptr != p){
		ptr = (unsigned short *)&blk;
		printf("WARNING!! freeing garbage (0x%x) pc = 0x%x %x\n",blk,
			ptr[-1],ptr[-2]);
		fflush(stdout);
		restore(i_state);
#ifdef	Z80
		printf("SP = 0x%x\n",getsp());
		for(;;) ;
#else
		return;
#endif
	}
	for(q = allocp; !(p > q && p < q->s.ptr); q = q->s.ptr){
		if (q >= q ->s.ptr && (p > q || p < q->s.ptr))
			break;
	}
	if(p + p->s.size == q->s.ptr){
		p->s.size += q->s.ptr->s.size;
		p->s.ptr = q->s.ptr->s.ptr;
	} else
		p->s.ptr = q->s.ptr;
	if(q + q->s.size == p){
		q->s.size += p->s.size;
		q->s.ptr = p->s.ptr;
	}
	else q->s.ptr = p;
	allocp = q;
	restore(i_state);
}

#ifdef	notdef
/* Move existing block to new area */
char *
realloc(area,size)
char *area;
unsigned size;
{
	unsigned osize;
	HEADER *hp;
	char *cp;
	char i_state;

	hp = ((HEADER *)area) - 1;
	osize = (hp->s.size -1) << 2;

	/* Make sure nobody else comes in and takes it */
	i_state = disable();
	free(area);
	if((cp = malloc(size)) != NULLCHAR && cp != area)
		memcpy((char *)cp,(char *)area,size>osize? osize : size);
	restore(i_state);
	return cp;
}
#endif
/* Allocate block of cleared memory */
char *
calloc(nelem,size)
unsigned nelem;	/* Number of elements */
unsigned size;	/* Size of each element */
{
	register unsigned i;
	register char *cp;
#ifdef	ALLOCDEBUG
	struct alloc *ap;
	short *ptr;
#endif
	i = nelem * size;
	if((cp = malloc(i)) != NULLCHAR)
		memset(cp,0,i);
#ifdef	ALLOCDEBUG
	if(cp == NULLCHAR)
		return cp;
	/* Adjust caller field to indicate real caller, not calloc() */
	ptr = (short *)&nelem;
	for(ap = alloc; ap < &alloc[NALLOC];ap++)
		if(ap->addr == cp)
			ap->pc = (long)ptr[-1];
#endif
	return cp;
}
/* Print free list map */
memstat()
{
	HEADER *p;
	unsigned total;

	printf("malloc fails %u\n",memfail);
	total = 0;
	p = &base;
	do {
		total += p->s.size * sizeof(HEADER);
		printf("0x%x %u",(unsigned)p,p->s.size * sizeof(HEADER));
		if(p == allocp)
			printf(" <-- allocp\n");
		else
			printf("\n");
		p = p->s.ptr;
	} while(p != &base);
	printf("Total: %u bytes\n",total);
#ifdef	ALLOCDEBUG
	{
		struct alloc *ap;
		printf("addr      size      caller\n");
		for(ap = alloc;ap < &alloc[NALLOC];ap++){
			if(ap->addr != NULLCHAR)
				printf("0x%-8x%-10d%-10lx\n",ap->addr,
					ap->size,ap->pc);
		}
	}
#endif
	fflush(stdout);
}
/* Grab as much memory as possible from the system and put it on the heap */
unsigned
grabcore()
{
	char *sbrk(),*cp;
	register HEADER *up;
	register unsigned int size;

	/* Initialize the heap pointers */
	if (allocp == NULLHDR){
		base.s.ptr = allocp = &base;
		base.s.size = 1;
	}
#ifdef MSC
#undef malloc
	size = _memavl() - 8400 ; /* leave some room for the stack to breath */
	cp = (char *)malloc(size) ;
#else
	/* Find out where the break is */
	cp = sbrk(0);
	/* Now try to push it as high as possible */
	for(size=256;;size += 256){
		if(brk(cp + size) == -1)
			break;
	}
#endif
#if defined(__TURBOC__)
	size -= 4096;
	brk(cp + size);
#endif
	up = (HEADER *)cp;
	up->s.size = size / sizeof(HEADER);
	up->s.ptr = up;	/* satisfy audit */
	free((char *)(up + 1));
	return size;
}

#ifdef MSC
/* If you add code hereafter, remember that for MSC, malloc has been undef'd
 * back to the original version which can't be used in an interrupt context.
 * Better to put the code before grabcore, nyet?
 */
#endif
